import re
import logging
import json
import os
from unittest.mock import AsyncMock, patch, mock_open
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.proximl as specimen

pytestmark = [mark.sdk, mark.unit]


@patch.dict(
    os.environ,
    {
        "PROXIML_USER": "user-id",
        "PROXIML_KEY": "key",
        "PROXIML_REGION": "region",
        "PROXIML_CLIENT_ID": "client_id",
        "PROXIML_POOL_ID": "pool_id",
        "PROXIML_API_URL": "api.example.com",
        "PROXIML_WS_URL": "api-ws.example.com",
    },
)
def test_proximl_from_envs():
    proximl = specimen.ProxiML()
    assert proximl.__dict__.get("api_url") == "api.example.com"
    assert proximl.__dict__.get("ws_url") == "api-ws.example.com"
    assert proximl.auth.__dict__.get("username") == "user-id"
    assert proximl.auth.__dict__.get("password") == "key"
    assert proximl.auth.__dict__.get("region") == "region"
    assert proximl.auth.__dict__.get("client_id") == "client_id"
    assert proximl.auth.__dict__.get("pool_id") == "pool_id"


def test_proximl_env_from_files():
    with patch(
        "proximl.proximl.open",
        mock_open(
            read_data=json.dumps(
                dict(
                    region="region_file",
                    client_id="client_id_file",
                    pool_id="pool_id_file",
                    api_url="api.example.com_file",
                    ws_url="api-ws.example.com_file",
                )
            )
        ),
    ):
        proximl = specimen.ProxiML()
    assert proximl.__dict__.get("api_url") == "api.example.com_file"
    assert proximl.__dict__.get("ws_url") == "api-ws.example.com_file"



import click
from proximl.cli import cli, pass_config, search_by_id_name
from proximl.cli.cloudbender import cloudbender


@cloudbender.group()
@pass_config
def datastore(config):
    """proxiML CloudBender datastore commands."""
    pass


@datastore.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to list datastores for.",
)
@pass_config
def list(config, provider, region):
    """List datastores."""
    data = [
        ["ID", "NAME", "TYPE", "URI", "ROOT"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    datastores = config.proximl.run(
        config.proximl.client.cloudbender.datastores.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    for datastore in datastores:
        data.append(
            [
                datastore.id,
                datastore.name,
                datastore.type,
                datastore.uri,
                datastore.root,
            ]
        )

    for row in data:
        click.echo(
            "{: >37.36} {: >29.28} {: >9.8} {: >12.11} {: >12.11}"
            "".format(*row),
            file=config.stdout,
        )


@datastore.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to create the datastore in.",
)
@click.option(
    "--type",
    "-t",
    type=click.Choice(
        [
            "nfs",
            "smb",
        ],
        case_sensitive=False,
    ),
    required=True,
    help="The type of datastore to create.",
)
@click.option(
    "--uri",
    "-u",
    type=click.STRING,
    required=True,
    help="The URI of the datastore",
)
@click.option(
    "--root",
    "-r",
    type=click.STRING,
    required=True,
    help="The root path to map within the datastore",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def create(config, provider, region, type, uri, root, name):
    """
    Creates a datastore.
    """
    return config.proximl.run(
        config.proximl.client.cloudbender.datastores.create(
            provider_uuid=provider,
            region_uuid=region,
            name=name,
            uri=uri,
            root=root,
            type=type,
        )
    )


@datastore.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to remove the datastore from.",
)
@click.argument("datastore", type=click.STRING)
@pass_config
def remove(config, provider, region, datastore):
    """
    Remove a datastore.

    DATASTORE may be specified by name or ID, but ID is preferred.
    """
    datastores = config.proximl.run(
        config.proximl.client.cloudbender.datastores.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    found = search_by_id_name(datastore, datastores)
    if None is found:
        raise click.UsageError("Cannot find specified datastore.")

    return config.proximl.run(found.remove())



import click
from proximl.cli import cli, pass_config, search_by_id_name


@cli.group()
@pass_config
def connection(config):
    """proxiML connection commands."""
    pass


@connection.command()
@pass_config
def list(config):
    """List connections."""
    data = [["ID", "TYPE", "STATUS"], ["-" * 80, "-" * 80, "-" * 80]]

    connections = config.proximl.run(config.proximl.client.connections.list())

    for con in connections:
        data.append([con.id, con.type, con.status])
    for row in data:
        click.echo(
            "{: >38.36} {: >9.7} {: >15.13}".format(*row), file=config.stdout
        )


@connection.command()
@click.argument("id", type=click.STRING)
@pass_config
def remove(config, id):
    """Remove connection."""
    connections = config.proximl.run(config.proximl.client.connections.list())

    found = search_by_id_name(id, connections)
    if None is found:
        raise click.UsageError("Connection ID specified does not exist.")

    if found.type == "dataset":
        this = config.proximl.run(config.proximl.client.datasets.get(id))
    elif found.type == "job":
        this = config.proximl.run(config.proximl.client.jobs.get(id))
    else:
        raise click.UsageError("Unknown connection type.")

    return config.proximl.run(this.disconnect())


@connection.command()
@click.option(
    "--all-projects/--no-all-projects",
    default=False,
    show_default=True,
    help="Auto attach to dataset and show creation logs.",
)
@pass_config
def remove_all(config, all_projects):
    """Clear and clean-up all proxiML connections."""
    return config.proximl.run(
        config.proximl.client.connections.remove_all(all_projects=all_projects)
    )



import re
from unittest.mock import AsyncMock
from pytest import mark, fixture

import proximl.environments as specimen

pytestmark = [mark.sdk, mark.unit, mark.environments]


@fixture
def environments(mock_proximl):
    yield specimen.Environments(mock_proximl)


@fixture
def environment(mock_proximl):
    yield specimen.Environment(
        mock_proximl,
        **{
            "id": "PYTORCH_PY38_17",
            "framework": "PyTorch",
            "py_version": "3.8",
            "version": "1.7",
            "cuda_version": "11.1",
            "name": "PyTorch 1.7 - Python 3.8",
        },
    )


class EnvironmentsTests:
    @mark.asyncio
    async def test_list_environments(self, environments, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await environments.list()
        mock_proximl._query.assert_called_once_with(
            f"/job/environments", "GET"
        )


class EnvironmentTests:
    def test_environment_properties(self, environment):
        assert isinstance(environment.id, str)
        assert isinstance(environment.name, str)
        assert isinstance(environment.py_version, str)
        assert isinstance(environment.framework, str)
        assert isinstance(environment.version, str)
        assert isinstance(environment.cuda_version, str)

    def test_environment_str(self, environment):
        string = str(environment)
        regex = r"^{.*\"id\": \"" + environment.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_environment_repr(self, environment):
        string = repr(environment)
        regex = (
            r"^Environment\( proximl , \*\*{.*'id': '"
            + environment.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import re
import json
import click
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises

pytestmark = [mark.cli, mark.unit, mark.jobs]

from proximl.cli import job as specimen
from proximl.jobs import Job


def test_list(runner, mock_jobs):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_proximl.jobs = AsyncMock()
        mock_proximl.jobs.list = AsyncMock(return_value=mock_jobs)
        result = runner.invoke(specimen, ["list"])
        print(result)
        assert result.exit_code == 0
        mock_proximl.jobs.list.assert_called_once()



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.device_configs as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.device_configs]


@fixture
def device_configs(mock_proximl):
    yield specimen.DeviceConfigs(mock_proximl)


@fixture
def device_config(mock_proximl):
    yield specimen.DeviceConfig(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        config_id="x",
        name="On-Prem DeviceConfig",
        model_uuid="model-id-1",
        image="nvidia/cuda",
        command="python run.py",
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_device_config(
        self,
        device_configs,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device_configs.get("1234", "5687", "91011")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/device/config/91011", "GET", {}
        )

    @mark.asyncio
    async def test_list_device_configs(
        self,
        device_configs,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device_configs.list("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/device/config", "GET", {}
        )

    @mark.asyncio
    async def test_remove_device_config(
        self,
        device_configs,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device_configs.remove("1234", "4567", "8910")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567/device/config/8910", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_device_config(self, device_configs, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            region_uuid="region-id-1",
            name="IoT 1",
            model_uuid="model-id-1",
            image="nvidia/cuda",
            command="python run.py",
        )
        expected_payload = dict(
            name="IoT 1",
            model_uuid="model-id-1",
            model_project_uuid="proj-id-1",
            image="nvidia/cuda",
            command="python run.py",
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "config_id": "device_config-id-1",
            "name": "IoT 1",
            "model_uuid": "model-id-1",
            "model_project_uuid": "proj-id-1",
            "image": "nvidia/cuda",
            "command": "python run.py",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await device_configs.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region/region-id-1/device/config",
            "POST",
            None,
            expected_payload,
        )
        assert response.id == "device_config-id-1"


class device_configTests:
    def test_device_config_properties(self, device_config):
        assert isinstance(device_config.id, str)
        assert isinstance(device_config.provider_uuid, str)
        assert isinstance(device_config.region_uuid, str)
        assert isinstance(device_config.name, str)

    def test_device_config_str(self, device_config):
        string = str(device_config)
        regex = r"^{.*\"config_id\": \"" + device_config.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_device_config_repr(self, device_config):
        string = repr(device_config)
        regex = (
            r"^DeviceConfig\( proximl , \*\*{.*'config_id': '"
            + device_config.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_device_config_bool(self, device_config, mock_proximl):
        empty_device_config = specimen.DeviceConfig(mock_proximl)
        assert bool(device_config)
        assert not bool(empty_device_config)

    @mark.asyncio
    async def test_device_config_remove(self, device_config, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device_config.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/device/config/x", "DELETE"
        )

    @mark.asyncio
    async def test_device_config_refresh(self, device_config, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "config_id": "device_config-id-1",
            "name": "IoT 1",
            "model_uuid": "model-id-1",
            "model_project_uuid": "proj-id-1",
            "image": "nvidia/cuda",
            "command": "python run.py",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await device_config.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a/device/config/x", "GET"
        )
        assert device_config.id == "device_config-id-1"
        assert response.id == "device_config-id-1"



import logging


class ProxiMLException(Exception):
    def __init__(self, message, *args):
        super().__init__(message, *args)
        self._message = message

    @property
    def message(self) -> str:
        return self._message

    def __repr__(self):
        return "ProxiMLException( {self.message!r})".format(self=self)

    def __str__(self):
        return "ProxiMLException({self.message!r})".format(self=self)


class ApiError(ProxiMLException):
    def __init__(self, status, data, *args):
        super().__init__(data, *args)
        self._status = status
        logging.debug(data)
        self._message = data.get("errorMessage") or data.get("message")

    @property
    def status(self) -> int:
        return self._status

    def __repr__(self):
        return "ApiError({self.status}, {self.message!r})".format(self=self)

    def __str__(self):
        return "ApiError({self.status}, {self.message!r})".format(self=self)


class JobError(ProxiMLException):
    def __init__(self, status, data, *args):
        super().__init__(data, *args)
        self._status = status
        self._message = data

    @property
    def status(self) -> str:
        return self._status

    def __repr__(self):
        return "JobError({self.status}, {self.message})".format(self=self)

    def __str__(self):
        return "JobError({self.status}, {self.message})".format(self=self)


class DatasetError(ProxiMLException):
    def __init__(self, status, data, *args):
        super().__init__(data, *args)
        self._status = status
        self._message = data

    @property
    def status(self) -> str:
        return self._status

    def __repr__(self):
        return "DatasetError({self.status}, {self.message})".format(self=self)

    def __str__(self):
        return "DatasetError({self.status}, {self.message})".format(self=self)


class ModelError(ProxiMLException):
    def __init__(self, status, data, *args):
        super().__init__(data, *args)
        self._status = status
        self._message = data

    @property
    def status(self) -> str:
        return self._status

    def __repr__(self):
        return "ModelError({self.status}, {self.message})".format(self=self)

    def __str__(self):
        return "ModelError({self.status}, {self.message})".format(self=self)


class CheckpointError(ProxiMLException):
    def __init__(self, status, data, *args):
        super().__init__(data, *args)
        self._status = status
        self._message = data

    @property
    def status(self) -> str:
        return self._status

    def __repr__(self):
        return "CheckpointError({self.status}, {self.message})".format(
            self=self
        )

    def __str__(self):
        return "CheckpointError({self.status}, {self.message})".format(
            self=self
        )


class ConnectionError(ProxiMLException):
    def __init__(self, message, *args):
        super().__init__(message, *args)
        self._message = message

    def __repr__(self):
        return "ConnectionError({self.message})".format(self=self)

    def __str__(self):
        return "ConnectionError({self.message})".format(self=self)


class SpecificationError(ProxiMLException):
    def __init__(self, attribute, message, *args):
        super().__init__(message, *args)
        self._attribute = attribute
        self._message = message

    @property
    def attribute(self) -> str:
        return self._attribute

    def __repr__(self):
        return "SpecificationError({self.attribute}, {self.message})".format(
            self=self
        )

    def __str__(self):
        return "SpecificationError({self.attribute}, {self.message})".format(
            self=self
        )



import click
from proximl.cli import cli, pass_config, search_by_id_name
from proximl.cli.cloudbender import cloudbender


@cloudbender.group()
@pass_config
def reservation(config):
    """proxiML CloudBender reservation commands."""
    pass


@reservation.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to list reservations for.",
)
@pass_config
def list(config, provider, region):
    """List reservations."""
    data = [
        ["ID", "NAME", "TYPE", "RESOURCE", "HOSTNAME"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    reservations = config.proximl.run(
        config.proximl.client.cloudbender.reservations.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    for reservation in reservations:
        data.append(
            [
                reservation.id,
                reservation.name,
                reservation.type,
                reservation.resource,
                reservation.hostname,
            ]
        )

    for row in data:
        click.echo(
            "{: >37.36} {: >29.28} {: >9.8} {: >9.8} {: >29.28}"
            "".format(*row),
            file=config.stdout,
        )


@reservation.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to create the reservation in.",
)
@click.option(
    "--type",
    "-t",
    type=click.Choice(
        [
            "port",
        ],
        case_sensitive=False,
    ),
    required=True,
    help="The type of reservation to create.",
)
@click.option(
    "--hostname",
    "-h",
    type=click.STRING,
    required=True,
    help="The hostname to make the reservation on",
)
@click.option(
    "--resource",
    "-r",
    type=click.STRING,
    required=True,
    help="The resource to reserve",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def create(config, provider, region, type, hostname, resource, name):
    """
    Creates a reservation.
    """
    return config.proximl.run(
        config.proximl.client.cloudbender.reservations.create(
            provider_uuid=provider,
            region_uuid=region,
            name=name,
            hostname=hostname,
            resource=resource,
            type=type,
        )
    )


@reservation.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to remove the reservation from.",
)
@click.argument("reservation", type=click.STRING)
@pass_config
def remove(config, provider, region, reservation):
    """
    Remove a reservation.

    RESERVATION may be specified by name or ID, but ID is preferred.
    """
    reservations = config.proximl.run(
        config.proximl.client.cloudbender.reservations.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    found = search_by_id_name(reservation, reservations)
    if None is found:
        raise click.UsageError("Cannot find specified reservation.")

    return config.proximl.run(found.remove())



import json
import logging


class Nodes(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, region_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node/{id}",
            "GET",
            kwargs,
        )
        return Node(self.proximl, **resp)

    async def list(self, provider_uuid, region_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node",
            "GET",
            kwargs,
        )
        nodes = [Node(self.proximl, **node) for node in resp]
        return nodes

    async def create(
        self,
        provider_uuid,
        region_uuid,
        friendly_name,
        hostname,
        minion_id,
        type="permanent",
        service="compute",
        **kwargs,
    ):
        logging.info(f"Creating Node {friendly_name}")
        data = dict(
            friendly_name=friendly_name,
            hostname=hostname,
            minion_id=minion_id,
            type=type,
            service=service,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node",
            "POST",
            None,
            payload,
        )
        node = Node(self.proximl, **resp)
        logging.info(f"Created Node {friendly_name} with id {node.id}")
        return node

    async def remove(self, provider_uuid, region_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node/{id}",
            "DELETE",
            kwargs,
        )


class Node:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._node = kwargs
        self._id = self._node.get("rig_uuid")
        self._provider_uuid = self._node.get("provider_uuid")
        self._region_uuid = self._node.get("region_uuid")
        self._type = self._node.get("type")
        self._service = self._node.get("service")
        self._name = self._node.get("friendly_name")
        self._hostname = self._node.get("hostname")
        self._status = self._node.get("status")
        self._online = self._node.get("online")
        self._maintenance_mode = self._node.get("maintenance_mode")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    @property
    def type(self) -> str:
        return self._type

    @property
    def service(self) -> str:
        return self._service

    @property
    def name(self) -> str:
        return self._name

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def status(self) -> str:
        return self._status

    @property
    def online(self) -> bool:
        return self._online

    @property
    def maintenance_mode(self) -> bool:
        return self._maintenance_mode

    def __str__(self):
        return json.dumps({k: v for k, v in self._node.items()})

    def __repr__(self):
        return f"Node( proximl , **{self._node.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}",
            "DELETE",
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self

    async def toggle_maintenance(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}/maintenance",
            "PATCH",
        )

    async def run_action(self, command):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}/action",
            "POST",
            None,
            dict(command=command),
        )



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.reservations as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.reservations]


@fixture
def reservations(mock_proximl):
    yield specimen.Reservations(mock_proximl)


@fixture
def reservation(mock_proximl):
    yield specimen.Reservation(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        reservation_id="x",
        name="On-Prem Reservation",
        type="port",
        resource="8001",
        hostname="service.local",
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_reservation(
        self,
        reservations,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await reservations.get("1234", "5687", "91011")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/reservation/91011", "GET", {}
        )

    @mark.asyncio
    async def test_list_reservations(
        self,
        reservations,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await reservations.list("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/reservation", "GET", {}
        )

    @mark.asyncio
    async def test_remove_reservation(
        self,
        reservations,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await reservations.remove("1234", "4567", "8910")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567/reservation/8910", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_reservation(self, reservations, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            region_uuid="region-id-1",
            name="On-Prem Reservation",
            type="port",
            resource="8001",
            hostname="service.local",
        )
        expected_payload = dict(
            name="On-Prem Reservation",
            type="port",
            resource="8001",
            hostname="service.local",
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "reservation_id": "reservation-id-1",
            "name": "On-Prem Reservation",
            "type": "port",
            "resource": "8001",
            "hostname": "service.local",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await reservations.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region/region-id-1/reservation",
            "POST",
            None,
            expected_payload,
        )
        assert response.id == "reservation-id-1"


class reservationTests:
    def test_reservation_properties(self, reservation):
        assert isinstance(reservation.id, str)
        assert isinstance(reservation.provider_uuid, str)
        assert isinstance(reservation.region_uuid, str)
        assert isinstance(reservation.type, str)
        assert isinstance(reservation.name, str)
        assert isinstance(reservation.resource, str)
        assert isinstance(reservation.hostname, str)

    def test_reservation_str(self, reservation):
        string = str(reservation)
        regex = r"^{.*\"reservation_id\": \"" + reservation.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_reservation_repr(self, reservation):
        string = repr(reservation)
        regex = (
            r"^Reservation\( proximl , \*\*{.*'reservation_id': '"
            + reservation.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_reservation_bool(self, reservation, mock_proximl):
        empty_reservation = specimen.Reservation(mock_proximl)
        assert bool(reservation)
        assert not bool(empty_reservation)

    @mark.asyncio
    async def test_reservation_remove(self, reservation, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await reservation.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/reservation/x", "DELETE"
        )

    @mark.asyncio
    async def test_reservation_refresh(self, reservation, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "reservation_id": "reservation-id-1",
            "name": "On-Prem Reservation",
            "type": "port",
            "resource": "8001",
            "hostname": "service.local",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await reservation.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a/reservation/x", "GET"
        )
        assert reservation.id == "reservation-id-1"
        assert response.id == "reservation-id-1"



import re
import json
import click
from unittest.mock import AsyncMock, patch, create_autospec
from pytest import mark, fixture, raises

pytestmark = [mark.cli, mark.unit, mark.projects]

from proximl.cli import project as specimen
from proximl.projects import Project


def test_list(runner, mock_projects):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_proximl.projects = AsyncMock()
        mock_proximl.projects.list = AsyncMock(return_value=mock_projects)
        result = runner.invoke(specimen, ["list"])
        print(result)
        assert result.exit_code == 0
        mock_proximl.projects.list.assert_called_once()


def test_list_datastores(runner, mock_project_datastores):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_project = create_autospec(Project)
        mock_project.list_datastores = AsyncMock(
            return_value=mock_project_datastores
        )
        mock_proximl.projects.get = AsyncMock(return_value=mock_project)
        result = runner.invoke(specimen, ["list-datastores"])
        print(result)
        assert result.exit_code == 0
        mock_project.list_datastores.assert_called_once()


def test_list_reservations(runner, mock_project_reservations):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_project = create_autospec(Project)
        mock_project.list_reservations = AsyncMock(
            return_value=mock_project_reservations
        )
        mock_proximl.projects.get = AsyncMock(return_value=mock_project)
        result = runner.invoke(specimen, ["list-reservations"])
        print(result)
        assert result.exit_code == 0
        mock_project.list_reservations.assert_called_once()



import click
from proximl.cli import cli, pass_config, search_by_id_name
from proximl.cli.cloudbender import cloudbender


@cloudbender.group()
@pass_config
def device(config):
    """proxiML CloudBender device commands."""
    pass


@device.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to list devices for.",
)
@pass_config
def list(config, provider, region):
    """List devices."""
    data = [
        [
            "ID",
            "NAME",
            "STATUS",
            "JOB STATUS",
            "ONLINE",
            "MAINTENANCE",
        ],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    devices = config.proximl.run(
        config.proximl.client.cloudbender.devices.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    for device in devices:
        data.append(
            [
                device.id,
                device.name,
                device.status,
                device.job_status,
                "X" if device.online else "",
                "X" if device.maintenance_mode else "",
            ]
        )

    for row in data:
        click.echo(
            "{: >37.36} {: >29.28} {: >9.8} {: >11.10} {: >7.6} {: >12.11}"
            "".format(*row),
            file=config.stdout,
        )


@device.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to create the region in.",
)
@click.option(
    "--minion-id",
    "-m",
    type=click.STRING,
    required=True,
    help="The minion_id of the new node.",
)
@click.option(
    "--hostname",
    "-h",
    type=click.STRING,
    help="The hostname (if different from name)",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def create(config, provider, region, minion_id, hostname, name):
    """
    Creates a node.
    """
    if not hostname:
        hostname = name
    return config.proximl.run(
        config.proximl.client.cloudbender.devices.create(
            provider_uuid=provider,
            region_uuid=region,
            friendly_name=name,
            hostname=hostname,
            minion_id=minion_id,
        )
    )


@device.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to delete the node from.",
)
@click.argument("device", type=click.STRING)
@pass_config
def remove(config, provider, region, node):
    """
    Remove a device.

    DEVICE may be specified by name or ID, but ID is preferred.
    """
    devices = config.proximl.run(
        config.proximl.client.cloudbender.devices.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    found = search_by_id_name(device, devices)
    if None is found:
        raise click.UsageError("Cannot find specified device.")

    return config.proximl.run(found.remove())



import json
import logging


class Reservations(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, region_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/reservation/{id}",
            "GET",
            kwargs,
        )
        return Reservation(self.proximl, **resp)

    async def list(self, provider_uuid, region_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/reservation",
            "GET",
            kwargs,
        )
        reservations = [
            Reservation(self.proximl, **reservation) for reservation in resp
        ]
        return reservations

    async def create(
        self,
        provider_uuid,
        region_uuid,
        name,
        type,
        resource,
        hostname,
        **kwargs,
    ):
        logging.info(f"Creating Reservation {name}")
        data = dict(
            name=name,
            type=type,
            resource=resource,
            hostname=hostname,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/reservation",
            "POST",
            None,
            payload,
        )
        reservation = Reservation(self.proximl, **resp)
        logging.info(f"Created Reservation {name} with id {reservation.id}")
        return reservation

    async def remove(self, provider_uuid, region_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/reservation/{id}",
            "DELETE",
            kwargs,
        )


class Reservation:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._reservation = kwargs
        self._id = self._reservation.get("reservation_id")
        self._provider_uuid = self._reservation.get("provider_uuid")
        self._region_uuid = self._reservation.get("region_uuid")
        self._type = self._reservation.get("type")
        self._name = self._reservation.get("name")
        self._resource = self._reservation.get("resource")
        self._hostname = self._reservation.get("hostname")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    @property
    def type(self) -> str:
        return self._type

    @property
    def name(self) -> str:
        return self._name

    @property
    def resource(self) -> str:
        return self._resource

    @property
    def hostname(self) -> str:
        return self._hostname

    def __str__(self):
        return json.dumps({k: v for k, v in self._reservation.items()})

    def __repr__(self):
        return f"Reservation( proximl , **{self._reservation.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/reservation/{self._id}",
            "DELETE",
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/reservation/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self



from .providers import Providers
from .regions import Regions
from .nodes import Nodes
from .devices import Devices
from .datastores import Datastores
from .reservations import Reservations
from .device_configs import DeviceConfigs


class Cloudbender(object):
    def __init__(self, proximl):
        self.proximl = proximl
        self.providers = Providers(proximl)
        self.regions = Regions(proximl)
        self.nodes = Nodes(proximl)
        self.devices = Devices(proximl)
        self.datastores = Datastores(proximl)
        self.reservations = Reservations(proximl)
        self.device_configs = DeviceConfigs(proximl)



import json
import logging


class Regions(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{id}", "GET", kwargs
        )
        return Region(self.proximl, **resp)

    async def list(self, provider_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region", "GET", kwargs
        )
        regions = [Region(self.proximl, **region) for region in resp]
        return regions

    async def create(self, provider_uuid, name, public, storage, **kwargs):
        logging.info(f"Creating Region {name}")
        data = dict(name=name, public=public, storage=storage, **kwargs)
        payload = payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region", "POST", None, payload
        )
        region = Region(self.proximl, **resp)
        logging.info(f"Created Region {name} with id {region.id}")
        return region

    async def remove(self, provider_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{id}", "DELETE", kwargs
        )


class Region:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._region = kwargs
        self._id = self._region.get("region_uuid")
        self._provider_uuid = self._region.get("provider_uuid")
        self._type = self._region.get("provider_type")
        self._name = self._region.get("name")
        self._status = self._region.get("status")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def type(self) -> str:
        return self._type

    @property
    def name(self) -> str:
        return self._name

    @property
    def status(self) -> str:
        return self._status

    def __str__(self):
        return json.dumps({k: v for k, v in self._region.items()})

    def __repr__(self):
        return f"Region( proximl , **{self._region.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}", "DELETE"
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self

    async def add_dataset(self, project_uuid, dataset_uuid, **kwargs):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}/dataset",
            "POST",
            None,
            dict(project_uuid=project_uuid, dataset_uuid=dataset_uuid),
        )

    async def add_model(self, project_uuid, model_uuid, **kwargs):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}/model",
            "POST",
            None,
            dict(project_uuid=project_uuid, model_uuid=model_uuid),
        )

    async def add_checkpoint(self, project_uuid, checkpoint_uuid, **kwargs):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}/checkpoint",
            "POST",
            None,
            dict(project_uuid=project_uuid, checkpoint_uuid=checkpoint_uuid),
        )



import json


class Environments(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def list(self):
        resp = await self.proximl._query(f"/job/environments", "GET")
        environments = [
            Environment(self.proximl, **environment) for environment in resp
        ]
        return environments


class Environment:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._environment = kwargs
        self._id = self._environment.get("id")
        self._name = self._environment.get("name")
        self._py_version = self._environment.get("py_version")
        self._framework = self._environment.get("framework")
        self._version = self._environment.get("version")
        self._cuda_version = self._environment.get("cuda_version")

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def py_version(self) -> str:
        return self._py_version

    @property
    def framework(self) -> str:
        return self._framework

    @property
    def version(self) -> str:
        return self._version

    @property
    def cuda_version(self) -> str:
        return self._cuda_version

    def __str__(self):
        return json.dumps(
            {k: v for k, v in self._environment.items() if k not in ["image"]}
        )

    def __repr__(self):
        return f"Environment( proximl , **{self._environment.__repr__()})"



import json
import asyncio
import math
import logging
import warnings
import webbrowser
from datetime import datetime

from proximl.exceptions import (
    ApiError,
    JobError,
    SpecificationError,
    ProxiMLException,
)
from proximl.connections import Connection


class Jobs(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, id, **kwargs):
        resp = await self.proximl._query(f"/job/{id}", "GET", kwargs)
        return Job(self.proximl, **resp)

    async def list(self, **kwargs):
        resp = await self.proximl._query(f"/job", "GET", kwargs)
        jobs = [Job(self.proximl, **job) for job in resp]
        return jobs

    async def create(
        self,
        name,
        type,
        gpu_type=None,
        gpu_types=[],
        gpu_count=None,
        cpu_count=None,
        disk_size=10,
        max_price=10,
        worker_commands=[],
        environment=dict(),
        data=dict(),
        model=dict(),
        endpoint=dict(),
        **kwargs,
    ):
        resources = {
            k: v
            for k, v in dict(
                gpu_count=gpu_count,
                disk_size=disk_size,
                max_price=max_price,
                cpu_count=cpu_count,
            ).items()
            if v is not None
        }

        if len(gpu_types):
            resources["gpu_types"] = gpu_types
        elif gpu_type:
            resources["gpu_type_id"] = gpu_type
        else:
            raise SpecificationError(
                "resources",
                "Invalid resource specification, either 'gpu_type' or 'gpu_types' must be provided",
            )

        config = dict(
            name=name,
            type=type,
            resources=resources,
            worker_commands=worker_commands,
            workers=kwargs.get("workers"),
            environment=environment,
            data=data,
            model=model,
            endpoint=endpoint,
            source_job_uuid=kwargs.get("source_job_uuid"),
            project_uuid=kwargs.get("project_uuid")
            or self.proximl.active_project,
        )
        payload = {
            k: v
            for k, v in config.items()
            if v or (k in ["model"] and not kwargs.get("source_job_uuid"))
        }
        if (
            not payload.get("worker_commands")
            and not payload.get("workers")
            and not kwargs.get("source_job_uuid")
        ):
            payload["worker_commands"] = []
        logging.info(f"Creating Job {name}")
        job = await self.create_json(payload)
        logging.info(f"Created Job {name} with id {job.id}")
        return job

    async def create_json(self, payload):
        logging.debug(f"Job payload: {payload}")
        resp = await self.proximl._query("/job", "POST", None, payload)
        job = Job(self.proximl, **resp)
        return job

    async def remove(self, id, **kwargs):
        await self.proximl._query(
            f"/job/{id}", "DELETE", dict(**kwargs, force=True)
        )


class Job:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._job = kwargs
        self._id = self._job.get("id", self._job.get("job_uuid"))
        self._name = self._job.get("name")
        self._status = self._job.get("status")
        self._type = self._job.get("type")
        self._workers = self._job.get("workers")
        self._credits = self._job.get("credits")
        self._project_uuid = self._job.get("project_uuid")

    @property
    def dict(self) -> dict:
        return {k: v for k, v in self._job.items() if k not in ["nb_token"]}

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def status(self) -> str:
        return self._status

    @property
    def type(self) -> str:
        return self._type

    @property
    def workers(self) -> list:
        return self._workers

    @property
    def credits(self) -> float:
        return self._credits

    @property
    def url(self) -> str:
        return self._job.get("endpoint").get("url")

    @property
    def notebook_url(self) -> str:
        if self.type != "notebook":
            return None
        return f"{self.url}/?token={self._job.get('nb_token')}"

    def __str__(self):
        return json.dumps(self.dict)

    def __repr__(self):
        return f"Job( proximl , {self.dict.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    def get_create_json(self):
        root_keys = [
            "name",
            "type",
            "resources",
            "model",
            "data",
            "environment",
            "endpoint",
            "workers",
            "project_uuid",
        ]
        resources_keys = [
            "gpu_count",
            "gpu_types",
            "disk_size",
            "max_price",
            "preemptible",
            "cpu_count",
        ]
        model_keys = [
            "source_type",
            "source_uri",
            "project_uuid",
            "checkpoints",
        ]
        data_keys = [
            "datasets",
            "input_type",
            "input_uri",
            "input_options",
            "output_type",
            "output_uri",
            "output_options",
        ]
        environment_keys = [
            "type",
            "env",
            "custom_image",
            "worker_key_types",
            "packages",
        ]
        endpoint_keys = ["routes", "start_command", "reservation_id"]
        create_json = dict()
        for k, v in self.dict.items():
            if k in root_keys:
                if k == "resources":
                    resources = dict()
                    for k2, v2 in v.items():
                        if k2 in resources_keys:
                            resources[k2] = v2
                    create_json["resources"] = resources
                elif k == "model":
                    model = dict()
                    for k2, v2 in v.items():
                        if k2 in model_keys:
                            model[k2] = v2
                    create_json["model"] = model
                elif k == "data":
                    data = dict()
                    for k2, v2 in v.items():
                        if k2 in data_keys:
                            data[k2] = v2
                    create_json["data"] = data
                elif k == "environment":
                    environment = dict()
                    for k2, v2 in v.items():
                        if k2 in environment_keys:
                            environment[k2] = v2
                    create_json["environment"] = environment
                elif k == "endpoint":
                    endpoint = dict()
                    for k2, v2 in v.items():
                        if k2 in endpoint_keys:
                            endpoint[k2] = v2
                    create_json["endpoint"] = endpoint
                elif k == "workers":
                    workers = []
                    for worker in v:
                        workers.append(worker.get("command"))
                    create_json["workers"] = workers
                else:
                    create_json[k] = v
        return create_json

    async def start(self):
        resp = await self.proximl._query(
            f"/job/{self._id}",
            "PATCH",
            dict(project_uuid=self._project_uuid),
            dict(command="start"),
        )
        self.__init__(self.proximl, **resp)
        return self

    async def stop(self):
        resp = await self.proximl._query(
            f"/job/{self._id}",
            "PATCH",
            dict(project_uuid=self._project_uuid),
            dict(command="stop"),
        )
        self.__init__(self.proximl, **resp)
        return self

    async def update(self, data):
        if self.type != "notebook":
            raise SpecificationError(
                "type",
                "Only notebook jobs can be modified.",
            )
        resp = await self.proximl._query(
            f"/job/{self._id}",
            "PATCH",
            dict(project_uuid=self._project_uuid),
            data,
        )
        self.__init__(self.proximl, **resp)
        return self

    async def get_worker_log_url(self, job_worker_uuid):
        resp = await self.proximl._query(
            f"/job/{self._id}/worker/{job_worker_uuid}/logs",
            "GET",
            dict(project_uuid=self._project_uuid),
        )
        return resp

    async def get_connection_utility_url(self):
        resp = await self.proximl._query(
            f"/job/{self._id}/download",
            "GET",
            dict(project_uuid=self._project_uuid),
        )
        return resp

    def get_connection_details(self):
        details = dict(
            entity_type="job",
            project_uuid=self._job.get("project_uuid"),
            cidr=self.dict.get("vpn").get("cidr"),
            ssh_port=self._job.get("vpn").get("client").get("ssh_port")
            if self._job.get("vpn").get("client")
            else None,
            model_path=self._job.get("model").get("source_uri")
            if self._job.get("model").get("source_type") == "local"
            else None,
            input_path=self._job.get("data").get("input_uri")
            if self._job.get("data").get("input_type") == "local"
            else None,
            output_path=self._job.get("data").get("output_uri")
            if self._job.get("data").get("output_type") == "local"
            else None,
        )
        return details

    async def open(self):
        if self.type != "notebook":
            raise SpecificationError(
                "type",
                "Only notebook jobs can be opened.",
            )
        webbrowser.open(self.notebook_url)

    async def connect(self):
        if self.type == "notebook" and self.status not in [
            "new",
            "waiting for data/model download",
            "waiting for GPUs",
            "waiting for resources",
        ]:
            raise SpecificationError(
                "type",
                "Notebooks cannot be connected to after model download is complete.  Use open() instead.",
            )
        if self.type == "endpoint" and self.status not in [
            "new",
            "waiting for data/model download",
        ]:
            return self.url
        if self.status in [
            "failed",
            "finished",
            "canceled",
            "archived",
            "removed",
            "removing",
        ]:
            raise SpecificationError(
                "status",
                f"You can only connect to active jobs.",
            )
        if self._job.get("vpn").get("status") == "n/a":
            logging.info("Local connection not enabled for this job.")
            return
        if self.status == "new":
            await self.wait_for("waiting for data/model download")
        connection = Connection(
            self.proximl, entity_type="job", id=self.id, entity=self
        )
        await connection.start()
        return connection.status

    async def disconnect(self):
        if self._job.get("vpn").get("status") == "n/a":
            logging.info("Local connection not enabled for this job.")
            return
        connection = Connection(
            self.proximl, entity_type="job", id=self.id, entity=self
        )
        await connection.stop()
        return connection.status

    async def remove(self, force=False):
        await self.proximl._query(
            f"/job/{self._id}",
            "DELETE",
            dict(project_uuid=self._project_uuid, force=force),
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/job/{self.id}", "GET", dict(project_uuid=self._project_uuid)
        )
        self.__init__(self.proximl, **resp)
        return self

    def _get_msg_handler(self, msg_handler):
        worker_numbers = {
            w.get("job_worker_uuid"): ind + 1
            for ind, w in enumerate(self._workers)
        }
        worker_numbers["data_worker"] = 0

        def handler(data):
            if data.get("type") == "subscription":
                data["worker_number"] = worker_numbers.get(data.get("stream"))
                if msg_handler:
                    msg_handler(data)
                else:
                    timestamp = datetime.fromtimestamp(
                        int(data.get("time")) / 1000
                    )
                    if len(self._workers) > 1:
                        print(
                            f"{timestamp.strftime('%m/%d/%Y, %H:%M:%S')}: Worker {data.get('worker_number')} - {data.get('msg').rstrip()}"
                        )
                    else:
                        print(
                            f"{timestamp.strftime('%m/%d/%Y, %H:%M:%S')}: {data.get('msg').rstrip()}"
                        )

        return handler

    async def attach(self, msg_handler=None):
        if (
            self.type == "notebook"
            and self.status != "waiting for data/model download"
        ):
            raise SpecificationError(
                "type",
                "Notebooks cannot be attached to after model download is complete.  Use open() instead.",
            )
        await self.refresh()
        if self.status not in ["finished", "failed"]:
            await self.proximl._ws_subscribe(
                "job",
                self._project_uuid,
                self.id,
                self._get_msg_handler(msg_handler),
            )

    async def copy(self, name, **kwargs):
        logging.debug(f"copy request - name: {name} ; kwargs: {kwargs}")
        if self.type != "notebook":
            raise SpecificationError(
                "job", "Only notebook job types can be copied"
            )

        job = await self.proximl.jobs.create(
            name,
            type=kwargs.get("type") or self.type,
            gpu_type=kwargs.get("gpu_type"),
            gpu_types=kwargs.get("gpu_types")
            or self._job.get("resources").get("gpu_types"),
            gpu_count=kwargs.get("gpu_count")
            or self._job.get("resources").get("gpu_count"),
            cpu_count=kwargs.get("cpu_count")
            or self._job.get("resources").get("cpu_count"),
            disk_size=kwargs.get("disk_size")
            or self._job.get("resources").get("disk_size"),
            max_price=kwargs.get("max_price")
            or self._job.get("resources").get("max_price"),
            worker_commands=kwargs.get("worker_commands"),
            workers=kwargs.get("workers"),
            environment=kwargs.get("environment"),
            data=kwargs.get("data"),
            model=kwargs.get("model"),
            source_job_uuid=self.id,
        )
        logging.debug(f"copy result: {job}")
        return job

    async def wait_for(self, status, timeout=300):
        valid_statuses = [
            "waiting for data/model download",
            "waiting for GPUs",
            "waiting for resources",
            "running",
            "stopped",
            "finished",
            "archived",
        ]
        if not status in valid_statuses:
            raise SpecificationError(
                "status",
                f"Invalid wait_for status {status}.  Valid statuses are: {valid_statuses}",
            )
        if status == "waiting for GPUs":
            warnings.warn(
                "'waiting for GPUs' status is deprecated, use 'waiting for resources' instead.",
                DeprecationWarning,
            )
        if (self.type == "training") and status == "stopped":
            warnings.warn(
                "'stopped' status is deprecated for training jobs, use 'finished' instead.",
                DeprecationWarning,
            )
        if self.status == status or (
            self.type == "training"
            and status == "finished"
            and self.status == "stopped"
        ):
            return

        POLL_INTERVAL_MIN = 5
        POLL_INTERVAL_MAX = 60
        POLL_INTERVAL = max(
            min(timeout / 60, POLL_INTERVAL_MAX), POLL_INTERVAL_MIN
        )
        retry_count = math.ceil(timeout / POLL_INTERVAL)
        count = 0
        while count < retry_count:
            await asyncio.sleep(POLL_INTERVAL)
            try:
                await self.refresh()
            except ApiError as e:
                if status == "archived" and e.status == 404:
                    return
                raise e
            if (
                self.status == status
                or (
                    self.type == "training"
                    and status == "finished"
                    and self.status == "stopped"
                )
                or (
                    status
                    in [
                        "waiting for GPUs",
                        "waiting for resources",
                    ]  ## this status could be very short and the polling could miss it
                    and self.status in ["starting", "provisioning", "running"]
                )
                or (
                    status
                    == "waiting for data/model download"  ## this status could be very short and the polling could miss it
                    and self.status in ["starting", "provisioning", "running"]
                )
            ):
                return self
            elif self.status == "failed":
                raise JobError(self.status, self)
            else:
                count += 1
                logging.debug(f"self: {self}, retry count {count}")

        raise ProxiMLException(f"Timeout waiting for {status}")
